use std::{
    fs::File,
    io::{BufRead, BufReader},
};

use hashbrown::HashMap;

struct StationMeasurements {
    min: f64,
    max: f64,
    temps: Vec<f64>,
}

const DEFAULT_HASHMAP_LENGTH: usize = 10000;

fn main() {
    let mut stations: HashMap<String, StationMeasurements> =
        HashMap::with_capacity(DEFAULT_HASHMAP_LENGTH);

    let file = File::open("../measurements.txt").expect("File measurements.txt not found");
    let reader = BufReader::new(file);
    let mut line_num = 0;
    for line_result in reader.lines() {
        line_num += 1;
        if line_num % 250000 == 0 {
            let formatted_line_num = format_nums(line_num);
            println!("Calculated {formatted_line_num} stations");
        }
        let line = line_result.expect("could not read line");
        let (station, temp) = line.split_once(';').unwrap();
        let temp = temp.parse().unwrap();
        let measurements_option = stations.get_mut(station);
        if let Some(measurements) = measurements_option {
            if temp < measurements.min {
                measurements.min = temp;
            } else if temp > measurements.max {
                measurements.max = temp;
            }
            measurements.temps.push(temp);
        } else {
            let measurements = StationMeasurements {
                min: temp,
                max: temp,
                temps: vec![temp],
            };
            stations.insert(station.to_owned(), measurements);
        }
    }
    for (station, measurments) in stations {
        let min = measurments.min;
        let max = measurments.max;
        let avg = avg(measurments.temps);
        println!("{station}={min}/{max}/{avg}");
    }
}

fn avg(temps: Vec<f64>) -> f64 {
    let num_temps = temps.len() as f64;
    let sum_temps: f64 = temps.iter().sum();
    sum_temps / num_temps
}

fn format_nums(num: usize) -> String {
    num.to_string()
        .as_bytes()
        .rchunks(3)
        .rev()
        .map(std::str::from_utf8)
        .collect::<Result<Vec<&str>, _>>()
        .unwrap()
        .join("_")
}
